/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <IL/il.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


//#define ENABLE_BMP /* not working */
//#define ENABLE_FITS /* not working */
//#define ENABLE_ILBM /* not working */
//#define ENABLE_JXR /* not working */

/*

  type         | Name                             | extansions
---------------+----------------------------------+---------------------------
* IL_BMP       | Bitmap                           | bmp                                | Not working (ilLoadL() failes)
  IL_CUT       | Dr. Halo                         | cut
  IL_DOOM      | Doom walls                       |
  IL_DOOM_FLAT | Doom flats                       |
  IL_ICO       | Icon                             | ico
  IL_JFIF      | JPEG/JFIF                        | jpg jpe jpeg
* IL_ILBM      | ILBM                             | iff, ilbm, lbm                     | Ofted not working (runs indefinitely)
  IL_PCD       | Kodak PhotoCD                    | pcd
* IL_PCX       | ZSoft Pcx                        | pcx
  IL_PIC       | Pic???                           | pic
  IL_PNM       | Portable Anymap                  | pbm, pgm, ppm
  IL_PNG       | Portable Network Graphics        | png
  IL_SGI       | Sgi                              | sgi bw rgb rgba
  IL_TGA       | Truevision Targa                 | tga
  IL_TIF       | Tiff                             | tif tiff
  IL_MDL       | Half-Life Model Texture          | mdl
  IL_WAL       | Quake 2 Texture                  | wal
  IL_LIF       | Homeworld Texture                | lif
* IL_MNG       | Multiple-image Network Graphics  | mng
  IL_JNG       |                                  | jng
* IL_GIF       | Graphics Interchange Format      | gif
* IL_DDS       | DirectDraw Surface               | dds
  IL_DCX       | ZSoft Multi-PCX                  | dcx
  IL_PSD       | Adobe PhotoShop                  | psd
  IL_PSP       | PaintShop Pro                    | psp
* IL_PIX       | PIX                              | pix
  IL_PXR       | Pixar                            | pxr
  IL_XPM       | X Pixel Map                      | xpm
  IL_HDR       | Radiance High Dynamic Range      | hdr
  IL_ICNS      | Macintosh Icon                   | icns
  IL_JP2       | Jpeg 2000                        | jp2
* IL_EXR       | OpenEXR                          | exr
* IL_WDP       | Microsoft HD Photo               | wdp, hdp                           | Not working (ilLoadL() failes)
  IL_VTF       | Valve Texture Format             | vtf                                |
  IL_WBMP      | Wireless Bitmap                  | wbmp                               |
  IL_SUN       | Sun Raster                       | sun, ras, rs, im1, im8, im24, im32 |
  IL_TPL       | Gamecube Texture                 | tpl
* IL_FITS      | Flexible Image Transport System  | fit, fits                          | Uses IL_LUMINANCE type
  IL_DICOM     | DICOM                            | dcm, dicom
  IL_IWI       | Call of Duty Infinity Ward Image | iwi
  IL_BLP       | Blizzard Texture Format          | blp
  IL_FTX       | Heavy Metal: FAKK2 Texture       | ftx
  IL_ROT       | Homeworld 2 - Relic Texture      | rot
  IL_TEXTURE   | Medieval II: Total War Texture   | texture
* IL_DPX       | Digital Picture Exchange         | dpx
  IL_UTX       | Unreal Texture                   | utx

*/

struct _abydos_plugin_handle_t{
    ILenum type;
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_devil_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-alias-pix",
#ifdef ENABLE_BMP
        "image/bmp",
#endif
        "image/x-digital-picture-exchange", "image/x-dpx",
        "image/x-direct-draw-surface", "image/x-dds",
        "image/x-exr",
#ifdef ENABLE_FITS
        "application/fits", "image/fits",
#endif
        "image/gif",
#ifdef ENABLE_ILBM
        "image/x-ilbm",
#endif
        "video/x-mng", "image/x-mng",
        "image/vnd.zbrush.pcx", "image/x-pcx",
#ifdef ENABLE_JXR
        "image/vnd.ms-photo",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_devil_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strcasecmp(mime_type,"image/x-alias-pix")){
        h->type = IL_PIX;
    }else
#if ENABLE_BMP
    if(!strcasecmp(mime_type,"image/bmp")){
        h->type = IL_BMP;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-digital-picture-exchange") ||
             !strcasecmp(mime_type,"image/x-dpx"))
    {
        h->type = IL_DPX;
    }else if(!strcasecmp(mime_type,"image/x-direct-draw-surface") ||
             !strcasecmp(mime_type,"image/x-dds"))
    {
        h->type = IL_DDS;
    }else if(!strcasecmp(mime_type,"image/x-exr")){
        h->type = IL_EXR;
    }else
#ifdef ENABLE_FITS
    if(!strcasecmp(mime_type,"application/fits") ||
       !strcasecmp(mime_type,"image/fits"))
    {
        h->type = IL_FITS;
    }else
#endif
    if(!strcasecmp(mime_type,"image/gif")){
        h->type = IL_GIF;
    }else
#ifdef ENABLE_ILBM
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->type = IL_ILBM;
    }else
#endif
    if(!strcasecmp(mime_type,"video/x-mng") ||
       !strcasecmp(mime_type,"image/x-mng"))
    {
        h->type = IL_MNG;
    }else if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx") ||
             !strcasecmp(mime_type,"image/x-pcx"))
    {
        h->type = IL_PCX;
    }else
#ifdef ENABLE_JXR
    if(!strcasecmp(mime_type,"image/vnd.ms-photo")){
        h->type = IL_WDP;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe_render = 1;
    h->surface = NULL;

    return h;
}


static void
_devil_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_interpret_palette_type(int type, int *r_bpc, nile_format_t *r_format)
{
    switch(type){
    case IL_PAL_RGB24:
        *r_bpc = 3;
        *r_format = NILE_FORMAT_r8g8b8;
        return 0;
    case IL_PAL_RGB32:
        *r_bpc = 4;
        *r_format = NILE_FORMAT_r8g8b8;
        return 0;
    case IL_PAL_RGBA32:
        *r_bpc = 4;
        *r_format = NILE_FORMAT_r8g8b8a8;
        return 0;
    case IL_PAL_BGR24:
        *r_bpc = 3;
        *r_format = NILE_FORMAT_b8g8r8;
        return 0;
    case IL_PAL_BGR32:
        *r_bpc = 4;
        *r_format = NILE_FORMAT_b8g8r8;
        return 0;
    case IL_PAL_BGRA32:
        *r_bpc = 4;
        *r_format = NILE_FORMAT_b8g8r8a8;
        return 0;
    }
    return -1;
}


#if 0
static uint32_t *
_palette_from_devil(ILubyte *src, int type, int colors){
    uint32_t *palette = (uint32_t*)malloc(256*sizeof(uint32_t));
    int i;
    if(colors > 256)
        colors = 256;
    switch(type){
    case IL_PAL_RGB24:
        for(i=0;i<colors;++i)
            palette[i] = src[i*3+0] << 16 | src[i*3+1] << 8 | src[i*3+2] << 0;
        break;
    case IL_PAL_RGB32:
        for(i=0;i<colors;++i)
            palette[i] = src[i*4+0] << 16 | src[i*4+1] << 8 | src[i*4+2] << 0;
        break;
    case IL_PAL_RGBA32:
        for(i=0;i<colors;++i){
            int alpha = src[i*4+3];
            palette[i] = (alpha*src[i*4+0]/255) << 16 |
                         (alpha*src[i*4+1]/255) <<  8 |
                         (alpha*src[i*4+2]/255) <<  0 |
                         alpha << 24;
        }
        break;
    case IL_PAL_BGR24:
        for(i=0;i<colors;++i)
            palette[i] = src[i*3+2] << 16 | src[i*3+1] << 8 | src[i*3+0] << 0;
        break;
    case IL_PAL_BGR32:
        for(i=0;i<colors;++i)
            palette[i] = src[i*4+2] << 16 | src[i*4+1] << 8 | src[i*4+0] << 0;
        break;
    case IL_PAL_BGRA32:
        for(i=0;i<colors;++i){
            int alpha = src[i*4+3];
            palette[i] = (alpha*src[i*4+2]/255) << 16 |
                         (alpha*src[i*4+1]/255) <<  8 |
                         (alpha*src[i*4+1]/255) <<  0 |
                         alpha << 24;
        }
        break;
    default:
        free(palette);
        return NULL;
    }
    for(i=colors;i<256;++i)
        palette[i] = 0;
    return palette;
}
#endif


static cairo_surface_t *
_surface_from_floats(const float *src, int width, int height){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    while(height-->0){
        int x;
        for(x=width;x>0;--x){
            float alpha = src[3];
            int r = 255*alpha*src[0];
            int g = 255*alpha*src[1];
            int b = 255*alpha*src[2];
            int a = 255*alpha;
            *dst++ = r<<16 | g<<8 | b<<0 | a<<24;
            src += 4;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_devil_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int depth;
    int format;
    ILubyte *src;
    if(!ilLoadL(h->type,(const void*)data,len))
        return -1;

    format = ilGetInteger(IL_IMAGE_FORMAT);
    DEBUG_p("format: %x",format);
    src = ilGetData();
    h->info->width = ilGetInteger(IL_IMAGE_WIDTH);
    h->info->height = ilGetInteger(IL_IMAGE_HEIGHT);
    depth = ilGetInteger(IL_IMAGE_BPP);
    DEBUG_p("depth: %d",depth);
    switch(format){
    case IL_COLOR_INDEX:
        if(depth != 1)
            break;
        {
            //uint32_t *palette;
            //palette = _palette_from_devil(ilGetPalette(),ilGetInteger(IL_PALETTE_TYPE),ilGetInteger(IL_PALETTE_NUM_COLS));
            uint8_t *palette = ilGetPalette();
            int colors = ilGetInteger(IL_PALETTE_NUM_COLS);
            int bpc;
            nile_format_t format;
            if(_interpret_palette_type(ilGetInteger(IL_PALETTE_TYPE),&bpc,&format)<0)
                break;
            h->surface = nil_cairo_surface_create_from_indexed(src,h->info->width,h->info->height,-1,palette,colors,bpc,format,0);
        }
        break;
    case IL_RGB:
        if(depth < 3 || depth > 4)
            break;
        h->surface = nil_cairo_surface_create_from_planar(src,h->info->width,h->info->height,-1,depth,NILE_FORMAT_r8g8b8,0);
        break;
    case IL_RGBA:
        switch(depth){
        case 4:
            h->surface = nil_cairo_surface_create_from_planar(src,h->info->width,h->info->height,-1,depth,NILE_FORMAT_r8g8b8a8,0);
            break;
        case 16:
            h->surface = _surface_from_floats((const float*)src,h->info->width,h->info->height);
            break;
        }
        break;
    case IL_BGR:
        if(depth < 3 || depth > 4)
            break;
        h->surface = nil_cairo_surface_create_from_planar(src,h->info->width,h->info->height,-1,depth,NILE_FORMAT_b8g8r8,0);
        break;
    case IL_BGRA:
        if(depth != 4 && depth != 16)
            break;
        h->surface = nil_cairo_surface_create_from_planar(src,h->info->width,h->info->height,-1,depth,NILE_FORMAT_b8g8r8a8,0);
        break;
    }
    ilDeleteImage(1);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_devil_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _devil_get_mime_types,
        .new = _devil_new,
        .free = _devil_free,
        .create_from_data = _devil_create_from_data,
        .get_image_surface = _devil_get_image_surface
        //.get_duration = _devil_get_duration
    };
    ilInit();
    return &p;
}
