/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>
#include <string.h>

#include "converter.h"


/*

Formats supported by farbfeld:

  avsff             | Stardent AVS X image
  bitff             | Super Famicom format
  bldff             | MegaPaint format
  bmpff             | Microsoft Windows bitmap image
  cgaff             | bsave - IBM PC Color Graphics Adapter
  cutff             | Dr.Halo .CUT
  drcsff            | DEC VT320 Dynamically Redefined Character Set
  faxff (faxff.ps)  | Postscript
  fhmff             | Free Hero Mesh
  fuunff            | Fuun RNA
  gemicnff          | GEM icons
  gifff             | Graphics Interchange Format
  gleff             | AMI BIOS Energy Star logo format
  grobff            | HP-48/HP-49 graphics object
  jefff             | Janome Embroidery Format
* jpegff            | Joint Photographic Experts Group
  kapff             | BSB/KAP nautical charts
  macff             | MacPaint
  makiff            | Maki-chan format
  mbff              | Hero Mesh puzzle set
  mcff              | Macrocell format
  mppff             | Multi Palette Picture
  mrfff             | Monochrome Recursive Format
  mzmff             | MegaZeux layer-mode MZM
  neoff             | NEOchrome
  pbmff             | Netpbm formats 1 to 7
  pmartff           | Atari ST Palette Master
* pngff             | Portable Network Graphics
  psff              | PostScript
  psytcff           | Psycopathicteen Tile Compressor
  rfntff            | RawFont
  rgbff             | Headerless true colour formats
  sinqlff           | Sinclair QL screen
  timaskff          | TeXnicard image mask
  vecff             | DRAWX vector graphics
  xbmff             | X bitmap image
  xorlff            | XORLINES-2
  xpmff             | X pixmap image
  xyzff             | RPG Maker 2000/2003 .XYZ picture
  xzipff            | Z-machine XZIP picture library
  zxff              | ZX Spectrum
  zztff             | ZZT world file or save game file

*/


static const char **
_ffconv_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef USE_JPG2FF
        "image/jpeg",
#endif
#ifdef USE_PNG2FF
        "image/png",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ffconv_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    const char *converter;
    h->info = info;
#ifdef USE_JPG2FF
    if(!strcasecmp(mime_type,"image/jpeg")){
        converter = "jpg2ff";
    }else
#endif
#ifdef USE_PNG2FF
    if(!strcasecmp(mime_type,"image/png")){
        converter = "png2ff";
    }else
#endif
    {
        free(h);
        return NULL;
    }
    if(!abydos_converter_init(h, converter, "image/x-farbfeld", 0)){
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    return h;
}


static void
_ffconv_free(abydos_plugin_handle_t *h)
{
    abydos_converter_done(h);
    free(h);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ffconv_get_mime_types,
        .new = _ffconv_new,
        .free = _ffconv_free,
        .progressive_begin = abydos_converter_begin,
        .progressive_feed = abydos_converter_feed,
        .progressive_end = abydos_converter_end,
        .get_image_surface = abydos_converter_get_image_surface
    };
    return &p;
}
