/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <debug.h>
#include <flif_dec.h>
#include <stdlib.h>                     /* malloc(), free() */


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_flif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-flif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_flif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_flif_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_img(abydos_plugin_handle_t *h, FLIF_IMAGE *img,int width,int height)
{
    uint8_t *row;
    cairo_surface_t *surface;
    uint32_t *dst;
    int stride;
    int y;
    row = (uint8_t*)malloc(width*4);
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    for(y=0;y<height;++y){
        int x;
        flif_image_read_row_RGBA8(img,y,row,width*4);
        for(x=0;x<width;++x){
            int r = row[x*4+0];
            int g = row[x*4+1];
            int b = row[x*4+2];
            int a = row[x*4+3];
            dst[x] = (a*r/255)<<16|
                     (a*g/255)<< 8|
                     (a*b/255)<< 0|
                     a<<24;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_flif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    FLIF_DECODER *dec;
    int res;
    int i;
    dec = flif_create_decoder();
    res = flif_decoder_decode_memory(dec,bytes,len);
    DEBUG_p("flif_decoder_decode_memory(): %d",res);

    h->info->frame_count = flif_decoder_num_images(dec);
    DEBUG_p("number of frames: %d",h->info->frame_count);

    h->frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        FLIF_IMAGE *img = flif_decoder_get_image(dec,i);
        int width = flif_image_get_width(img);
        int height = flif_image_get_height(img);
        uint32_t delay;
        h->frame[i].surface = _surface_from_img(h,img,width,height);
        delay = flif_image_get_frame_delay(img);
        //DEBUG_p("frame %d delay: %d",i,(int)delay);
        h->frame[i].duration = (double)delay/1000;
        if(width>h->info->width)
            h->info->width = width;
        if(height>h->info->height)
            h->info->height = height;
    }

    return 0;
}


static cairo_surface_t *
_flif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_flif_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _flif_get_mime_types,
        .new = _flif_new,
        .free = _flif_free,
        .create_from_data = _flif_create_from_data,
        .get_image_surface = _flif_get_image_surface,
        .get_duration = _flif_get_duration
    };
    return &p;
}
