/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include "debug.h"
#include "nil-atarist.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


void
nil_atarist_copy_colors(uint8_t *dst, int bpc, nile_format_t format, const uint8_t *src, int count)
{
    int i;
    if(NILE_FORMAT_r5g6b5 == format){
        for(i=0;i<count;++i){
            uint16_t col = be16toh(*(uint16_t*)(src+i*2));
            int r = (col>>8)&0x07;
            int g = (col>>4)&0x07;
            int b = (col>>0)&0x07;
            r = (r<<2) | (r>>1);
            g = (g<<3) | (g<<0);
            b = (b<<2) | (b>>1);
            *(uint16_t*)dst = r<<1 | g<<5 | b<<0;
            dst += 2;
            src += 2;
        }
    }else{
        int r_o = NILE_OFFSET_RED(format);
        int g_o = NILE_OFFSET_GREEN(format);
        int b_o = NILE_OFFSET_BLUE(format);
        for(i=0;i<count;++i){
            uint16_t col = be16toh(*(uint16_t*)src);
            int r = (col>>8)&0x07;
            int g = (col>>4)&0x07;
            int b = (col>>0)&0x07;
            dst[r_o] = r<<5 | r<<2 | r>>1;
            dst[g_o] = g<<5 | g<<2 | g>>1;
            dst[b_o] = b<<5 | b<<2 | b>>1;
            dst += bpc;
            src += 2;
        }
    }
}


void
nil_atarist_bitplanes_from_vram(nile_t *ni, int resolution, const uint8_t *src, const uint8_t *palette)
{
    int width;
    int height;
    int depth;
    int deep;
    switch(resolution){
    case 0:
        width = 320;
        height = 200;
        depth = 4;
        deep = 1;
        break;
    case 1:
        width = 640;
        height = 200;
        depth = 2;
        deep = 1;
        break;
    case 2:
        width = 640;
        height = 400;
        depth = 1;
        deep = 0;
        break;
    }
    nile_ensure_bitplanes(ni, width, height, depth, deep);
    if(deep){
        int stride = nile_get_bitplanes_stride(ni);
        int offs = 0;
        int y;
        for(y=0;y<height;++y){
            int x;
            for(x=0;x<width/16;++x){
                int plane;
                for(plane=0;plane<depth;++plane){
                    ni->bitplanes[plane][offs + x*2+0] = *src++;
                    ni->bitplanes[plane][offs + x*2+1] = *src++;
                }
            }
            offs += stride;
        }
    }else{
        int plane;
        for(plane=0;plane<depth;++plane){
            ni->bitplanes[plane] = (uint8_t*)src;
            src += width/8 * height;
        }
    }
    nile_ensure_palette(ni, 1<<depth, ni->bpc);
    nil_atarist_copy_colors(ni->palette, ni->bpc, ni->format, palette, 1<<depth);
}


int nil_atarist_vram_encode(nil_atarist_vram_t *vram, const nile_t *ni){
    const uint8_t *src = ni->pixels;
    int src_stride = nile_get_stride(ni);
    uint8_t *dst = vram->screen;
    int dst_stride;
    int y;
    int i;
    DEBUG_p("image size: %dx%d",ni->width,ni->height);
    switch(vram->resolution[1]){
    case 0:
        dst_stride = ni->width/2;
        for(y=0;y<ni->height;++y){
            int x;
            for(x=0;x<ni->width/16;++x){
                int plane;
                for(plane=0;plane<4;++plane){
                    uint8_t *b = dst+x*2*4+plane*2;
                    b[0] = 0;
                    b[1] = 0;
                    for(i=0;i<8;++i){
                        if(src[x*16+0+i] & (1<<plane))
                            b[0] |= 128>>i;
                        if(src[x*16+8+i] & (1<<plane))
                            b[1] |= 128>>i;
                    }
                }
            }
            src += src_stride;
            dst += dst_stride;
        }
        break;
    case 1:
        dst_stride = ni->width/4;
        for(y=0;y<ni->height;++y){
            int x;
            for(x=0;x<ni->width/16;++x){
                int plane;
                for(plane=0;plane<2;++plane){
                    uint8_t *b = dst+x*2*2+plane*2;
                    b[0] = 0;
                    b[1] = 0;
                    for(i=0;i<8;++i){
                        if(src[x*16+0+i] & (1<<plane))
                            b[0] |= 128>>i;
                        if(src[x*16+8+i] & (1<<plane))
                            b[1] |= 128>>i;
                    }
                }
            }
            src += src_stride;
            dst += dst_stride;
        }
        break;
    default:
        return -1;
    }

    for(i=0;i<MIN(16,ni->colors);++i){
        int r = ni->palette[i*ni->bpc+0]>>5;
        int g = ni->palette[i*ni->bpc+1]>>5;
        int b = ni->palette[i*ni->bpc+2]>>5;
        int rgb = r<<8|g<<4|b<<0;
        vram->palette[i*2+0] = rgb >> 8;
        vram->palette[i*2+1] = rgb & 255;
    }
    for(i=ni->colors;i<16;++i){
        vram->palette[i*2+0] = 0;
        vram->palette[i*2+1] = 0;
    }
    return 0;
}
